unit Menus;

interface
uses
  Windows, OpenGL, HUD, TextureManager, SoundManager;
  
type
  TCustomMenuItem = class
  public
    Caption : String;
    Parent : TObject;
    procedure OnSelect; virtual; abstract;
  end;

  TNewGameMenuItem = class (TCustomMenuItem)
    procedure OnSelect; override;
  end;

  TOptionsMenuItem = class (TCustomMenuItem)
    procedure OnSelect; override;
  end;

  TQuitMenuItem = class (TCustomMenuItem)
    procedure OnSelect; override;
  end;

  TResumeMenuItem = class (TCustomMenuItem)
    procedure OnSelect; override;
  end;

  TBackMenuItem = class (TCustomMenuItem)
    procedure OnSelect; override;
  end;

  TVideoSetupMI = class(TCustomMenuItem)
    procedure OnSelect; override;
  end;

  TSoundSetupMI = class(TCustomMenuItem)
    procedure OnSelect; override;
  end;

  TControllerSetupMI = class(TCustomMenuItem)
    procedure OnSelect; override;
  end;

  TCustomMenu = class
  private
    Items : array of TCustomMenuItem;
    MainFont : TSystemFont;
    X, Y, Selected : Integer;
    Bg : Integer;
    Up : TCustomMenu;
    MouseSelected : Boolean;
  public
    procedure Render; virtual;
    procedure KeyDown(Key : Byte); virtual;
    procedure KeyUp(Key : Byte); virtual; abstract;
    constructor Create(tX,tY : Integer; Parent : TCustomMenu); virtual;
    procedure Resume; virtual;
    procedure AddElement(Caption : String; Item : TCustomMenuItem; OnTop : Boolean = False);
    procedure MouseMove(mX, mY : Integer); virtual;
    procedure MouseDown(mX, mY : Integer); virtual;
  end;

  TOptionsMenu = class (TCustomMenu)
  public
    constructor Create(tX,tY : Integer; Parent : TCustomMenu); override;
  end;

  TMainMenu = class (TCustomMenu)
  private
    OptionsMenu : TOptionsMenu;
  public
    constructor Create(tX,tY : Integer; Parent : TCustomMenu); override;
    procedure KeyDown(Key : Byte); override;
  end;

Var
  ActiveMenu : TCustomMenu;
  MenuSound : Integer;

implementation
uses
  TheGame, u_winapi, g_global_vars, MissionStart;

{ TNewGameMenuItem }

procedure TNewGameMenuItem.OnSelect;
begin
  if not ((Parent as TCustomMenu).Items[0] is TResumeMenuItem) then
    (Parent as TCustomMenu).AddElement('Resume', TResumeMenuItem.Create, True);
    
  SoundMngr.StopAll;
  MissionBrief := TMissionBrief.Create;
  MissionBrief.Text := 'Mission 1:'#13#10'Training for the next mission';
  GameMode := modeMission;
end;

{ TCustomMenu }

constructor TCustomMenu.Create(tX, tY: Integer; Parent : TCustomMenu);
begin
  X := tX;
  Y := tY;
  MainFont := TSystemFont.Create('Fixedsys', 32);
  Selected := 0;
  Up := Parent;
end;

procedure TCustomMenu.KeyDown(Key: Byte);
begin
  if Key = vk_down then
  begin
    inc(Selected);
    if Selected > High(Items) then
      Selected := 0;
  end else
  if Key = vk_up then
  begin
    dec(Selected);
    if Selected < 0 then
      Selected := High(Items);
  end else
  if Key = vk_Return then
  begin
    Items[Selected].OnSelect;
  end else
  if Key = VK_ESCAPE then
  begin
    if Up <> nil then
      ActiveMenu := UP; 
  end;
end;

procedure TCustomMenu.Render;
var
  I: Integer;
begin
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity;

  gluOrtho2D(0, Screen_Width, Screen_Height, 0);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;

  glEnable(GL_TEXTURE_2D);
  TextureMngr.Bind(Bg);
  glColor3f(1, 1, 1);
  glBegin(GL_QUADS);
    glTexCoord2f(0, 1); glVertex2i(0, 0);
    glTexCoord2f(1, 1); glVertex2i(Screen_Width, 0);
    glTexCoord2f(1, 0); glVertex2i(Screen_Width, Screen_Height);
    glTexCoord2f(0, 0); glVertex2i(0, Screen_Height);
  glEnd;

  for i := 0 to High(Items) do
  begin
    if Selected = i then
      MainFont.Print(X, Y + I * 30, '> ' + Items[I].Caption, 255, 0, 0) else
        MainFont.Print(X, Y + I * 30, '  ' + Items[I].Caption, 0, 0, 0);
  end;
end;

procedure TCustomMenu.Resume;
begin
  SoundMngr.Play(MenuSound);
  glDisable(GL_DEPTH_TEST);
end;

procedure TCustomMenu.MouseMove(mX, mY : Integer);
var
  s:Integer;
begin
  MouseSelected := False;
  mY := mY+100;
  if (mY > Y) and (mY < Y + (Length(Items) * 30)) then
  begin
    s := (mY - Y) div 30;
    if (mX > X) and (mX < X + (Length(Items[s].Caption) * 30)) then
    begin
      Selected := s;
      MouseSelected := True;
    end;
  end;
end;

procedure TCustomMenu.MouseDown(mX, mY: Integer);
begin
  MouseMove(mX, mY);
  if MouseSelected then
    Items[Selected].OnSelect;
end;

procedure TCustomMenu.AddElement(Caption: String; Item: TCustomMenuItem;
  OnTop: Boolean);
var
  Ind : Integer;
begin
  SetLength(Items, Length(Items)+1);
  if OnTop then
  begin
    for Ind := High(Items) downto 0 do
      Items[Ind] := Items[Ind - 1];

    Ind := 0;
  end else
    Ind := High(Items);

  Items[Ind] := Item;
  Items[Ind].Caption := Caption;
  Items[Ind].Parent := Self;
end;


{ TMainMenu }

constructor TMainMenu.Create(tX, tY: Integer; Parent : TCustomMenu);
begin
  inherited;
  AddElement('New Game', TNewGameMenuItem.Create);
  AddElement('Options', TOptionsMenuItem.Create);
  AddElement('Quit', TQuitMenuItem.Create);

  Bg := TextureMngr.LoadTexture('data\textures\gui\menu_bg.jpg', True);
  OptionsMenu := TOptionsMenu.Create(tX, tY, Self);
  OptionsMenu.Bg := Bg;
end;

procedure TMainMenu.KeyDown(Key: byte);
begin
  inherited;
end;

{ TQuitMenuItem }

procedure TQuitMenuItem.OnSelect;
begin
  inherited;
  Quit_Hit := True;
end;

{ TOptionsMenuItem }

procedure TOptionsMenuItem.OnSelect;
begin
  inherited;
  ActiveMenu := (Parent as TMainMenu).OptionsMenu;
end;

{ TResumeMenuItem }

procedure TResumeMenuItem.OnSelect;
begin
  inherited;
  SoundMngr.PauseAll;
  GameMode := modeGame;
  Game.Resume;
end;

{ TOptionsMenu }

constructor TOptionsMenu.Create(tX, tY: Integer; Parent : TCustomMenu);
begin
  inherited;
  //AddElement('Video', TVideoSetupMI.Create);
  //AddElement('Sound', TSoundSetupMI.Create);
  //AddElement('Controllers', TControllerSetupMI.Create);
  AddElement('<< Back', TBackMenuItem.Create);
end;

{ TBackMenuItem }

procedure TBackMenuItem.OnSelect;
begin
  inherited;
  ActiveMenu := (Parent as TCustomMenu).Up;
end;

{ TControllerSetupMI }

procedure TControllerSetupMI.OnSelect;
begin
  inherited;

end;

{ TVideoSetupMI }

procedure TVideoSetupMI.OnSelect;
begin
  inherited;

end;

{ TSoundSetupMI }

procedure TSoundSetupMI.OnSelect;
begin
  inherited;

end;

end.
