{
  Texture Manager
  by Kefir (kefir.rwr@gmail.com)
  25 aug 2009
}
unit TextureManager;

interface
uses
  Windows, OpenGL, JPEG, Classes, Graphics;

const
  ERROR_TITLE = 'Texture Manager Error';

  TGA_NO_IMAGE = 0;
  TGA_PALETTE = 1;
  TGA_TRUE_COLOR = 2;
  TGA_MONOCHROME = 3;
  TGA_PALETTE_COMPRESSED = 9;
  TGA_TRUE_COLOR_COMPRESSED = 10;
  TGA_MONOCHROME_COMPRESSED = 11;


type
  TTextureRec  = record
      Name : String;
      ID : GLuint;
      Width, Height : Integer;
      BPP : Byte;
  end;

  TTextureManager = class
  private
    function AddTexture(Name : String; ID : GLuint; Width, Height : Integer; BPP : Byte):Integer;
    function CreateTexture(Width, Height : Integer; BPP : Byte; Mode : GLuint; Data : Pointer):GLuint;
    function LoadBMP(const FileName : String): GLuint;
    function LoadTGA(const FileName : String): GLuint;
    function LoadJPEG(const FileName: String): GLuint;
    procedure Error(Mess : string);
  public
    Textures : array of TTextureRec;
    RGBTransparency : Boolean;
    TransparencyColor : record
      r,g,b:Byte;
    end;
    Smooth : Boolean;
    ErrorHandler : procedure (ErrMessage : String);
    LogHandler : procedure (Mess : String);
    function LoadTexture(const FileName : String; tSmooth : Boolean = false) : GLuint;
    procedure SetTransparentRGBColor(red,green,blue : Byte);
    procedure Bind(ID : Integer);
    constructor Create;
  end;

  TGA_Header = packed record   // Header type for TGA images
    FileType       : Byte;
    ColorMapType   : Byte;
    ImageType      : Byte;
    ColorMapStart  : WORD;
    ColorMapLength : WORD;
    ColorMapDepth  : Byte;
    OrigX          : WORD;
    OrigY          : WORD;
    Width          : WORD;
    Height         : WORD;
    BPP            : Byte;
    ImageInfo      : Byte;
  end;

var
  TextureMngr: TTextureManager;

procedure glBindTexture(target: GLenum; texture: GLuint); stdcall; external opengl32;
function gluBuild2DMipmaps(Target: GLenum; Components, Width, Height: GLint; Format, atype: GLenum; Data: Pointer): GLint; stdcall; external glu32;
procedure glGenTextures(n: GLsizei; var textures: GLuint); stdcall; external opengl32;

implementation

type
  TRGBArr = array[0..0] of array[0..2] of Byte;
  TRGBAArr = array[0..0] of array[0..3] of Byte;

{ TTextureManager }

function TTextureManager.AddTexture(Name: String; ID: GLuint; Width,
  Height: Integer; BPP: Byte): Integer;
begin
  SetLength(Textures, Length(Textures)+1);

  Result := High(Textures);
  Textures[Result].Name := Name;
  Textures[Result].ID := ID;
  Textures[Result].Width := Width;
  Textures[Result].Height := Height;
  Textures[Result].BPP := BPP;
end;

procedure TTextureManager.Bind(ID: Integer);
begin
  glBindTexture(GL_TEXTURE_2D, Textures[ID].ID);
end;

constructor TTextureManager.Create;
begin
  RGBTransparency := False;
  Smooth := False;
end;

function TTextureManager.CreateTexture(Width, Height: Integer; BPP: Byte;
  Mode: GLuint; Data: Pointer): GLuint;
var
  Err : Integer;
begin
  glGenTextures(1, Result);
  glBindTexture(GL_TEXTURE_2D, Result);

  if Smooth then
  begin
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
  end else
  begin
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_NEAREST_MIPMAP_NEAREST);
  end;

  Err := gluBuild2DMipmaps(GL_TEXTURE_2D, BPP div 8, Width, Height, Mode, GL_UNSIGNED_BYTE, Data);

  if Err  <> 0 then
  begin
    Error(gluErrorString(Err));
  end;
end;

procedure TTextureManager.Error(Mess: string);
begin
  if @ErrorHandler <> nil then
    ErrorHandler(Mess) else
      MessageBox(0, PChar(Mess), ERROR_TITLE, MB_ICONERROR);
end;

function TTextureManager.LoadBMP(const FileName: String): GLuint;
var
  bmpfh : TBitmapFileHeader;
  bmpih : TBitmapInfoHeader;
  BmpFile : File of Byte;
  srcData, dstData : Pointer;
  t : Byte;
  i, bpp, mode:Integer;
begin
  Result := 0;
  
  AssignFile(BmpFile, FileName);
  Reset(BmpFile);

  BlockRead(BmpFile, bmpfh, Sizeof(bmpfh));
  BlockRead(BmpFile, bmpih, Sizeof(bmpih));

  if bmpih.biCompression = BI_RGB then
  begin
    if (bmpih.biBitCount = 24) then
    begin
      if RGBTransparency then
      begin
        bpp := 32;
        mode := GL_RGBA;
      end else
      begin
        bpp := 24;
        mode := GL_RGB;
      end;
          
      GetMem(srcData, bmpih.biWidth*bmpih.biHeight*(bmpih.biBitCount div 8));
      BlockRead(BmpFile, srcData^, bmpih.biWidth*bmpih.biHeight*(bmpih.biBitCount div 8));

      for i := 0 to (bmpih.biWidth*bmpih.biHeight)-1 do
      begin
        t := TRGBArr(srcData^)[i][0];
        TRGBArr(srcData^)[i][0] := TRGBArr(srcData^)[i][2];
        TRGBArr(srcData^)[i][2] := t;
      end;

      if RGBTransparency then
      begin
        GetMem(dstData, bmpih.biWidth*bmpih.biHeight*4);
        for i := 0 to (bmpih.biWidth*bmpih.biHeight)-1 do
        begin
          CopyMemory(@TRGBAArr(dstData^)[i], @TRGBArr(srcData^)[i], 3);
        
          if (TRGBAArr(dstData^)[i][0] = TransparencyColor.r) and (TRGBAArr(dstData^)[i][1] = TransparencyColor.g) and (TRGBAArr(dstData^)[i][2] = TransparencyColor.b) then
            TRGBAArr(dstData^)[i][3] := 0 else
              TRGBAArr(dstData^)[i][3] := 255;
        end;
      end else
        dstData := srcData;

      Result := CreateTexture(bmpih.biWidth, bmpih.biHeight, bpp, mode, dstData);
      Result := AddTexture(FileName, Result, bmpih.biWidth, bmpih.biHeight, bmpih.biBitCount);

      if dstData = srcData then
        FreeMem(dstData) else
        begin
          FreeMem(dstData);
          FreeMem(srcData);
        end;

    end else
      Error('Unsupported bitmap bitcount in: ' + FileName);
  end else
    Error('Compressed bitmaps not supported: ' + FileName);

  CloseFile(BmpFile);
end;

{------------------------------------------------------------------}
{  Load JPEG textures  by Jan Horn                                 }
{------------------------------------------------------------------}
function TTextureManager.LoadJPEG(const FileName: String): GLuint;
var
  Data : Array of LongWord;
  W, Width : Integer;
  H, Height : Integer;
  BMP : TBitmap;
  JPG : TJPEGImage;
  C : LongWord;
  Line : ^LongWord;
begin
  JPG:=TJPEGImage.Create;
  try
    JPG.LoadFromFile(Filename);
  except
    Error('Error while loading JPEG: ' + FileName);
    Result := 0;
    Exit;
  end;

  // Create Bitmap
  BMP:=TBitmap.Create;
  BMP.pixelformat:=pf32bit;
  BMP.width:=JPG.width;
  BMP.height:=JPG.height;
  BMP.canvas.draw(0,0,JPG);        // Copy the JPEG onto the Bitmap

  //  BMP.SaveToFile('D:\test.bmp');
  Width :=BMP.Width;
  Height :=BMP.Height;
  SetLength(Data, Width*Height);

  For H:=0 to Height-1 do
  Begin
    Line :=BMP.scanline[Height-H-1];   // flip JPEG
    For W:=0 to Width-1 do
    Begin
      c:=Line^ and $FFFFFF; // Need to do a color swap
      Data[W+(H*Width)] :=(((c and $FF) shl 16)+(c shr 16)+(c and $FF00)) or $FF000000;  // 4 channel.
      inc(Line);
    End;
  End;

  Result := CreateTexture(jpg.Width, jpg.Height, 32, GL_RGBA, Data);
  Result := AddTexture(FileName, Result, bmp.Width, bmp.Height, 24);

  BMP.free;
  JPG.free;
  //Freemem(Data);
end;

function TTextureManager.LoadTexture(const FileName: String; tSmooth : Boolean = false): GLuint;
  function LowerCase(const S: string): string;
  var
    Ch: Char;
    L: Integer;
    Source, Dest: PChar;
  begin
    L := Length(S);
    SetLength(Result, L);
    Source := Pointer(S);
    Dest := Pointer(Result);
    while L <> 0 do
    begin
      Ch := Source^;
      if (Ch >= 'A') and (Ch <= 'Z') then Inc(Ch, 32);
      Dest^ := Ch;
      Inc(Source);
      Inc(Dest);
      Dec(L);
    end;
  end;
var
  ext : String;
  i   : Integer;
begin
  Smooth := tSmooth;
  Result := 0;

  for i := 0 to High(Textures) do
  begin
    if Textures[i].Name = FileName then
    begin
      Result := i;
      Exit;
    end;
  end;
  
  if Pos('.', FileName) <> 0 then
  begin
    ext := FileName;
    Delete(ext, 1, Pos('.', ext));

    ext := LowerCase(ext);

    if ext = 'bmp' then
      Result := LoadBMP(FileName) else
    if ext = 'tga' then
      Result := LoadTGA(FileName) else
    if (ext = 'jpg') or (ext = 'jpeg') then
      Result := LoadJPEG(FileName) else
        Error('Unsupported file format (' + ext + '): ' + FileName);

    if @LogHandler <> nil then
      LogHandler('Texture Manager: ' + FileName); 

  end else
    Error('File have no extension. Can''''t recognize format');
end;

function TTextureManager.LoadTGA(const FileName: String): GLuint;
var
  TgaFile : File of Byte;
  TgaHeader : TGA_Header;
  Data : Pointer;
  FileId : array[0..255] of char;
  Mode, I : Integer;
  t : Byte;
begin

  Result := 0;

  AssignFile(TgaFile, FileName);
  Reset(TgaFile);

  BlockRead(TgaFile, TgaHeader, SizeOf(TGA_Header));

  if (TgaHeader.ImageType = TGA_TRUE_COLOR) or (TgaHeader.ImageType = TGA_MONOCHROME) then
  begin
    if TgaHeader.ImageType = TGA_MONOCHROME then
      Mode := GL_RED else
    begin
      if TgaHeader.BPP = 24 then
        Mode := GL_RGB else
          Mode := GL_RGBA;
    end;

    BlockRead(TgaFile, FileId, TgaHeader.FileType);

    GetMem(Data, TgaHeader.Width * TgaHeader.Height * (TgaHeader.BPP div 8));

    BlockRead(TgaFile, Data^, TgaHeader.Width * TgaHeader.Height * (TgaHeader.BPP div 8));

    for I := 0 to (TgaHeader.Width * TgaHeader.Height)-1 do
    begin
      if TgaHeader.BPP = 24 then
      begin
        t := TRGBArr(Data^)[I][0];
        TRGBArr(Data^)[I][0] := TRGBArr(Data^)[I][2];
        TRGBArr(Data^)[I][2] := t;
      end else if TgaHeader.BPP = 32 then
      begin
        t := TRGBAArr(Data^)[I][0];
        TRGBAArr(Data^)[I][0] := TRGBAArr(Data^)[I][2];
        TRGBAArr(Data^)[I][2] := t;
      end else Break;
    end;

    Result := CreateTexture(TgaHeader.Width, TgaHeader.Height, TgaHeader.BPP, Mode, Data);
    Result := AddTexture(FileName, Result, TgaHeader.Width, TgaHeader.Height, TgaHeader.BPP);

    FreeMem(Data);
  end else
    Error('Unsupported TGA format'); 

  CloseFile(TgaFile);
end;

procedure TTextureManager.SetTransparentRGBColor(red, green, blue: Byte);
begin
  RGBTransparency := True;
  TransparencyColor.r := red;
  TransparencyColor.g := green;
  TransparencyColor.b := blue;
end;

end.
