unit TheGame;

interface
uses
  Windows, OpenGL, u_opengl, glCamera, GLTypes, GameTypes, u_winapi, HUD, TextureManager, PLayer, SoundManager, g_global_vars, uAmbience, uGameObjects, Sheduler;

type
  TGameController = class
  public
    procedure Move; virtual; abstract;
    procedure SwitchTo; virtual; abstract;
  end;
  
  TGame = class (TGameObject)
  private
    Ambience : TAmbience;
    
    CameraMode : (cmFree, cmPlayer, cmTPS);
    Pause : Boolean;
  public
    sndMainLoop,
    sndEngine, // data\sfx\engine_loop.ogg
    sndMorse,
    sndMorse2,
    sndAim,
    sndMissile : Integer;
    sndFlyBy : Integer;
    Camera : TGLCamera;
    Player : TPlayer;
    Objects : array of TGameObject;
    Controller : TGameController;
    HUD : THUD;
    Font : TSystemFont;
    Messag : String;
    procedure KeyDown(Key : Byte);
    function AddObject(Obj : TGameObject; First : Boolean = False): Integer;
    procedure DeleteObject(id : Integer);
    procedure Render; override;
    procedure Move; override;
    procedure DoCollision(CollisionedWith : TObject); override;
    procedure Resume;
    procedure ResetObjects;
    procedure SwitchController(Cont : TGameController);
    function GetNearest(Obj : TGameObject) : TGameObject;
    function CountObjecstByType(Typ : String): Integer;
    constructor Create(MissileTyp : TMissileType);
  end;

  TBasicGameController = class (TGameController)
  private
    Parent : TGame;
  public
    constructor Create(Father : TGameObject); virtual;
  end;

  TControllerFight = class (TBasicGameController)
    TimeTo, TTLTime : Integer;
    procedure Move; Override;
    procedure SwitchTo; override;
  end;

  TControllerRefueling = class (TBasicGameController)
  private
    Stage : DWORD;
  public
    KC135 : TC135;
    procedure Move; override;
    procedure SwitchTo; override;
    constructor Create(Father : TGameObject); override;
  end;

  TControllerLandining = class (TBasicGameController)
  private
    Stage : DWORD;
    Carrier : TCarrier;
  public
    procedure Move; override;
   procedure SwitchTo; override;
  end;

  TControllerTakeOff = class (TBasicGameController)
  private
    Carrier : TGameObject;
    sndTakeOff : Integer;
    Stage, Fin : DWORD;
  public
    procedure Move; override;
    procedure SwitchTo; override;
    constructor Create(Father : TGameObject); override;
  end;
var
  Game : TGame;

implementation
uses
  Menus, Math;

{ TGame }

function TGame.AddObject(Obj: TGameObject; First : Boolean): Integer;
var
  i:Integer;
begin
  SetLength(Objects, Length(Objects)+1);

  if First then
  begin
    for i := High(Objects) downto 1 do
      Objects[i] := Objects[i-1];

    Objects[0] := Obj;
  end else
  begin
    Objects[High(Objects)] := Obj;
  end;
  Result := High(Objects);
end;

function TGame.CountObjecstByType(Typ: String): Integer;
var
  i : Integer;
begin
  Result := 0;

  for i := 0 to High(Objects) do
    if Objects[i].TypeName = Typ then inc(Result);
end;

constructor TGame.Create(MissileTyp : TMissileType);
begin
  Camera := TGLCamera.Create;
  Camera.SetPosition(0, 5, 0);
  Camera.SetTarget(0, 5, 1);

  HUD := THUD.Create;
  Font := TSystemFont.Create('Fixedsys', 24);

  Player := TPlayer.Create(MissileTyp);
  Player.Alt := 80;

  sndMainLoop := SoundMngr.LoadSound('data\music\game_loop.ogg', True);
  sndEngine := SoundMngr.LoadSound('data\sfx\engine_loop.ogg', True);
  sndMorse := SoundMngr.LoadSound('data\sfx\radio\morse.ogg');
  sndMorse2 := SoundMngr.LoadSound('data\sfx\radio\morse2.ogg', True);
  sndFlyBy := SoundMngr.LoadSound('data\sfx\flyby.ogg');
  sndAim := SoundMngr.LoadSound('data\sfx\radio\aim.ogg');
  sndMissile := SoundMngr.LoadSound('data\sfx\missile_start.ogg');
  //SoundMngr.Play(sndMainLoop);

  Ambience := TAmbience.Create(abDay, Camera);

  Controller := TControllerTakeOff.Create(Self);
  //Controller := TControllerRefueling.Create(Self);
  Controller.SwitchTo;

  CameraMode := cmPlayer;
end;

procedure TGame.DeleteObject(id: Integer);
begin
  Objects[id].Free;
  Objects[id] := Objects[High(Objects)];
  SetLength(Objects, length(Objects)-1);
end;

procedure TGame.DoCollision(CollisionedWith: TObject);
var
  i,j : Integer;
begin
  for i := 0 to High(Objects)-1 do
  begin
    if Objects[i].Collision then
    begin
      for j := i+1 to High(Objects) do
      begin
        if Objects[j].Collision then
        begin
          if Distance(Objects[i].Pos, Objects[j].Pos) <= Objects[i].Radius+Objects[j].Radius then
          begin
            Objects[i].DoCollision(Objects[j]);
            Objects[j].DoCollision(Objects[i]);
          end;
        end;
      end;
    end;
  end;
end;

function TGame.GetNearest(Obj : TGameObject): TGameObject;
var
  i : Integer;
  minH, d : Single;
begin
  Result := nil;
  minH := 123456789.;
  for i := 0 to High(Objects) do
  begin
    if (Objects[i] <> Obj) and (Objects[i].Trigger1 = False) and (Objects[i].TypeName = 'Mig 28') then
    begin
      d := Distance(Obj.Pos, Objects[i].Pos);
      if d <= minH then
      begin
        Result := Objects[i];
        minH := d;
      end;
    end;
  end;
end;

procedure TGame.KeyDown(Key: Byte);
begin
  if Key = vk_escape then
  begin
    GameMode := modeMenu;
    SoundMngr.PauseAll;
    ActiveMenu.Resume;
  end else
  if Key = VK_F2 then
  begin
    if Player.Fuel < 20 then
    begin
      SwitchController(TControllerRefueling.Create(Self));
    end;
  end else
  if Key = VK_SPACE then
  begin
    if Player.AimTo = nil then
    begin
      Player.AimTo := GetNearest(Player);
      if Player.AimTo <> nil then
      begin
        Player.AimTo.Trigger1 := True;
        SoundMngr.Play(sndAim);
      end;
    end else
    begin
      Player.FireRocket;
      SoundMngr.Play(sndMissile);
      Player.AimTo := nil;
    end;
  end;{ else
  if Key = VK_F4 then
  begin
    PushController;
    SwitchController(TControllerLandining.Create(Self));
  end; }
end;

procedure TGame.Move;
var
  i:Integer;
begin
  if not Pause then
  begin

    DoCollision(Self);
    
    for i := High(Objects) downto 0 do
    begin
      if Objects[i].MustDie then DeleteObject(i);
    end;

    for i := High(Objects) downto 0 do
    begin
      Objects[i].Move;
    end;

    Controller.Move;
  end;

  Camera.Move;

  case CameraMode of
//cmFree   : Camera.DebugMove(Keys, Mouse_Dx, Mouse_Dy);
    cmPlayer : begin
                 Camera.Position := Player.Pos;
                 Camera.Target := AddVectors(Player.Pos, Player.Direction);
                 Camera.Bank := Player.Bank;
               end;
    cmTPS     : begin
                  Camera.Target := Player.Pos;
                  Camera.Position := AddVectors(Player.Pos, MultyplyVectorScalar(Player.Direction, -2));
                  Camera.Bank := 0;//Player.Bank;
                end;
  end;

  if Not Pause then
  begin
    Ambience.Move;
    HUD.Move;
  end;
end;

procedure TGame.Render;
var
  i:Integer;
begin
  Camera.Look;
  glPushMatrix;
  glTranslatef(0, 2, 0);
  glPopMatrix;
  Ambience.Render;
  
  for i := 0 to High(Objects) do
  begin
    Objects[i].Render;
    Objects[i].Project;
  end;

  
  if CameraMode = cmPlayer then
  begin
    Hud.Render;
  end else
    Player.Render;

  Hud.SetOrtho;
  Font.Print((Screen_Width div 2) - (Length(Messag) * 30) div 2, 200, Messag, 255, 128);

end;

procedure TGame.ResetObjects;
var
  i:Integer;
begin
  for i := 0 to High(Objects) do
    Objects[i].Free;

  SetLength(Objects, 0);

end;

procedure TGame.Resume;
begin
  SoundMngr.Play(sndMainLoop);
  glEnable(GL_DEPTH_TEST);
end;

procedure TGame.SwitchController(Cont: TGameController);
begin

  if Assigned(Controller) then
    Controller.Free;

  Controller := nil;

  Controller := Cont;
  Controller.SwitchTo;
end;

{ TControllerTakeOff }

constructor TControllerTakeOff.Create;
begin
  inherited;
  sndTakeOff :=  SoundMngr.LoadSound('data\sfx\takeoff.ogg');
end;

procedure TControllerTakeOff.Move;
begin
  inherited;

  if Stage <> 0 then
        Parent.Player.Move;;

  if Stage = 1 then
  begin
    Parent.Player.Speed := Parent.Player.Speed + 0.001;
    if Parent.Player.Pos.Z <= -6 then
      Stage := 2;
  end else
  if Stage = 2 then
  begin
    Parent.Player.Tilt := Parent.Player.Tilt - 0.2;
    Parent.Player.Alt := Parent.Player.Alt - (Sin(Parent.Player.Tilt / (180/pi)) * 0.1);
    Parent.Player.Speed := Parent.Player.Speed + 0.0005;
    if Parent.Player.Tilt <= -60 then
      Parent.Player.Tilt := -60;

    if Parent.Player.Alt >= 35 then
      Stage := 3;
  end else
  if Stage = 3 then
  begin
    Parent.Player.Tilt := Parent.Player.Tilt + 0.2;
    Parent.Player.Alt := Parent.Player.Alt - (Sin(Parent.Player.Tilt / (180/pi)) * 0.1);
    if Parent.Player.Tilt >= 0 then
    begin
      Parent.Player.Tilt := 0;
      Shedulr.SetVarAfter(@Fin, 1, stDword, 1000);
    end;
  end;

  if Fin = 1 then
  begin
    Parent.SwitchController(TControllerFight.Create(Parent));
    Carrier.MustDie := True;
  end;
end;

procedure TControllerTakeOff.SwitchTo;
begin
  inherited;
  Parent.HUD.Info[1] := 'Take off!!';
  Parent.ResetObjects;
  Carrier := TCarrier.Create;
  Carrier.Pos := CreateVector(0, 0, 0);
  Parent.AddObject(Carrier);
  Parent.Player.HAngle := 180;
  Parent.Player.Pos := CreateVector(-0.5, 1.5, 6.8);
  Parent.Player.Alt := 1.5;
  Parent.Player.Speed := 0.05;
  
  SoundMngr.Play(sndTakeOff);
  Shedulr.SetVarAfter(@Stage, 1, stWord, 9000);
end;

{ TControllerFight }

procedure TControllerFight.Move;
var
  i:Integer;
  at : TMigAttackType;
begin
  inherited;
  Parent.Player.Move;

  inc(TimeTo);

  if Parent.Player.Fuel < 20 then
  begin
    if odd(GetTickCount div 500) then
      Parent.HUD.Info[1] := 'LOW FUEL! PRESS  [ F2 ]!' else
        Parent.HUD.Info[1] := '';
  end;

  if TimeTo = 500 then
  begin
    at := TMigAttackType(Random(3));
    for i := 0 to 2 do
    begin
      Parent.AddObject(TMig28.Create(Parent.Player, at, i));
    end;
    TimeTo := 0;
  end;

  if TTLTime = 15000 then
    Parent.SwitchController(TControllerLandining.Create(Parent));

  inc(TTLTime);

end;

procedure TControllerFight.SwitchTo;
begin
  inherited;
  Parent.HUD.Info[0] := '';
  Parent.HUD.Info[1] := '';

  SoundMngr.Play(Parent.sndMainLoop);
  SoundMngr.Play(Parent.sndEngine);
  Parent.Player.KeyControll := True;
  Parent.Player.EnableBank := True;
  Parent.HUD.Radar := TBasicRadar.Create;
//  Parent.AddObject(TMig28.Create(Parent.Player, atFlyTo));
end;

{ TControllerRefueling }

constructor TControllerRefueling.Create(Father: TGameObject);
begin
  inherited;
  KC135 := TC135.Create(Parent.Player);
end;

procedure TControllerRefueling.Move;
var
  BackFight : TControllerFight;
begin
  Parent.Player.Move;
  if Stage = 0 then
  begin
    if Parent.CountObjecstByType('Mig 28') = 0 then
    begin
      Stage := 1;
      Parent.Player.KeyControll := False;
      Parent.ResetObjects;
      SoundMngr.StopAll;
      SoundMngr.Play(Parent.sndEngine);
      SoundMngr.Play(Parent.sndMorse2);
    end;
  end else
  if Stage = 1 then
  begin
    if Parent.Player.Alt < 80 then
    begin
      if Parent.Player.Tilt > -30 then
        Parent.Player.Tilt := Parent.Player.Tilt - 0.5;

      if Parent.Player.Bank > 0 then Parent.Player.Bank := Parent.Player.Bank - 0.5;
      if Parent.Player.Bank < 0 then Parent.Player.Bank := Parent.Player.Bank + 0.5;

      Parent.Player.Alt := Parent.Player.Alt - (Sin(Parent.Player.Tilt / (180/pi)) * 0.1);

      if Parent.Player.Speed > 2 then
        Parent.Player.Speed := Parent.Player.Speed - 0.001;
    end else
      Stage := 2;
  end else
  if Stage = 2 then
  begin
    Parent.AddObject(KC135);
    Stage := 3;
  end else
  if Stage = 3 then
  begin
    if Parent.Player.Tilt < 0 then Parent.Player.Tilt := Parent.Player.Tilt + 0.7;
    if KC135.Shlang >= 15 then
    begin
      Stage := 5;
      SoundMngr.Stop(Parent.sndMorse2);
    end;
  end else
  if Stage = 4 then
  begin
    //  
  end else
  if Stage = 5 then
  begin
    //  
    if Parent.Player.Fuel < 100 then
      Parent.Player.Fuel := Parent.Player.Fuel + 0.1 else
      begin
        KC135.Away := True;
        Stage := 6;
      end;
  end else
  if Stage = 6 then
  begin
    if Distance(Parent.Player.Pos, KC135.Pos) >= 80 then
    begin
      KC135.MustDie := True;
      Stage := 7;
    end;
  end else
  if Stage = 7 then
  begin
      if Parent.Player.Tilt < 60 then Parent.Player.Tilt  := Parent.Player.Tilt  + 0.3;
    Parent.Player.Alt := Parent.Player.Alt - (Sin(Parent.Player.Tilt / (180/pi)) * 0.1);
    if Parent.Player.Alt <= 50 then
    begin
      Parent.Player.KeyControll := True;
      Parent.Player.EnableBank := True;
      Parent.Player.AimTo := nil;
      //Parent.PopController;
      BackFight := TControllerFight.Create(Parent);
      BackFight.TTLTime := 10000;

      Parent.SwitchController(BackFight);
      SoundMngr.Play(Parent.sndMainLoop);
    end;
  end;

end;

procedure TControllerRefueling.SwitchTo;
begin
  inherited;
  SoundMngr.Stop(Parent.sndMainLoop);
  SoundMngr.Play(Parent.sndMorse);
end;

{ TControllerLandining }

procedure TControllerLandining.Move;
begin
  Parent.Player.Move;
  if Stage = 0 then
  begin
    if Parent.CountObjecstByType('Mig 28') = 0 then
      Stage := 1;
  end else
  if Stage = 1 then
  begin
    Parent.Player.Bank := 0;
    Parent.Player.KeyControll := False;
    Stage := 2;

  end else
  if Stage = 2 then
  begin
    if Parent.Player.Tilt < 12 then Parent.Player.Tilt := Parent.Player.Tilt + 0.1;
    //if Parent.Player.Tilt > 12 then Parent.Player.Tilt := Parent.Player.Tilt - 0.1;
    
    if Parent.Player.Alt > 7 then Parent.Player.Alt := Parent.Player.Alt - 0.1;
    if Parent.Player.Speed > 0.15 then Parent.Player.Speed := Parent.Player.Speed - 0.001;

    if (Parent.Player.Tilt >= 12) and (Parent.Player.Alt <= 7) and (Parent.Player.Speed <= 0.15) then
    begin
      Carrier := TCarrier.Create;
      Carrier.Pos := AddVectors(Parent.Player.Pos, MultyplyVectorScalar(Parent.Player.Direction, 800));
      Carrier.Pos.Y := 0;
      Carrier.HAngle := -Parent.Player.HAngle;

      Parent.HUD.Radar := TLendRadar.Create;

      Parent.HUD.Info[0] := 'Speed: 288';
      Parent.HUD.Info[1] := 'Tilt: 10';

      Parent.AddObject(Carrier);
      Parent.Camera.cfSpeed := 0.1;
      Parent.Camera.Fov := 50;
      Stage := 3;
    end;
  end else
  if Stage = 3 then
  begin
    Parent.Player.KeyControll := True;
    Parent.Player.EnableBank := False;
    Parent.Player.NormTilt := 20;
    Parent.Player.SpeedUpTilt := False;
    Parent.Player.BackTilt := False;
    Parent.Player.TiltLimit := 20;

    Stage := 4;
  end else
  if Stage = 4 then
  begin

    if Keys[ord('Z')] then
      Parent.Player.Speed := Parent.Player.Speed + (Sin(Parent.Player.Tilt / (180/pi)) * 0.001);

    if Parent.Player.Speed < 0.1 then
      Parent.Player.Speed := 0.1;

    if Parent.Player.Speed > 0.5 then
      Parent.Player.Speed := 0.5;

    Parent.Player.Tilt := Parent.Player.Tilt + (-100 + Random(200))/700;

    if Distance(Parent.Player.Pos, Carrier.Pos) < 100 then
    begin
      Parent.Player.KeyControll := False;
      Stage := 5;
    end;
  end else
  if Stage = 5 then
  begin
    if (Parent.Player.Speed > 0.250) and (Parent.Player.Speed < 0.295) then
      Parent.Player.Speed := 0.288;

    if (Parent.Player.Tilt > 8) and (Parent.Player.Tilt < 12) then
      Parent.Player.Tilt := 10;

    Parent.CameraMode := cmFree;
    Parent.Camera.Position := AddVectors( Carrier.Pos, MultyplyVectorScalar( Parent.Player.Direction, 15));
    Parent.Camera.Position.Y := Parent.Camera.Position.Y + 5;
    Parent.Camera.ActualFov := 3;
    Parent.Camera.cfSpeed := 0.05;
    Parent.Camera.Fov := 35;
    Stage := 6;
  end else
  if Stage = 6 then
  begin
    Parent.Camera.Target := Parent.Player.Pos;
    if Distance(CreateVector(Parent.Player.Pos.X, 0, Parent.Player.Pos.Z), Carrier.Pos) > 7 then
      Parent.Player.Alt := Parent.Player.Alt - (Sin(Parent.Player.Tilt / (180/pi)) * 0.1) else
    if (Parent.Player.Alt < 3) and (Parent.Player.Alt > 1) then
    begin
      Parent.Player.Alt := 1.5;
      if Parent.Player.Speed > 0 then Parent.Player.Speed := Parent.Player.Speed - 0.0035;
      if Parent.Player.Speed < 0 then
      begin
        Parent.Player.Speed := 0;
        Stage := 100;
        Shedulr.SetVarAfter(@Stage, 7, stDword, 5000);
        Parent.Camera.Fov := 60;
        SoundMngr.StopAll;
        SoundMngr.Play(SoundMngr.LoadSound('data\music\topgun_anthem.ogg', True));
        Parent.Messag := 'You Win!!! 8)';
      end;
    end else

    if Parent.Player.Alt < 1.5 then
    begin
      // FAIL
      Parent.Messag := 'You Lose!!! 8)';
      SoundMngr.Play(SoundMngr.LoadSound('data\sfx\impact.ogg', False));
      Stage := 8;
    end;
  end else
  if Stage = 7 then
  begin
    Parent.Camera.Position.X := Parent.Player.Pos.X + Sin(GetTickCount / 10000) * 8;
    Parent.Camera.Position.Z := Parent.Player.Pos.Z + Cos(GetTickCount / 10000) * 8;
    Parent.Camera.Position.Y := Parent.Player.Pos.Y + 5;
  end else
  if Stage = 8 then
  begin

  end;
end;

procedure TControllerLandining.SwitchTo;
begin
  inherited;
    SoundMngr.Stop(Parent.sndMainLoop);
    SoundMngr.Play(Parent.sndMorse);
    Parent.HUD.Info[1] := 'To landining sequence';
end;

{ TBasicGameController }

constructor TBasicGameController.Create(Father: TGameObject);
begin
  Parent := (Father as TGame);
end;

end.
